/*
 * mdb.c - MINIX program debugger
 *
 * Written by Bruce D. Szablak
 *
 * This free software is provided for non-commerical use. No warrantee
 * of fitness for any use is implied. You get what you pay for. Anyone
 * may make modifications and distribute them, but please keep this header
 * in the distribution.
 */

#define  MAIN
#include "mdb.h"
#include <sys/wait.h>
#include <ctype.h>
#include <sgtty.h>
#include <signal.h>

#define MAXARG	50
#define MAXB	10
#define MAXLINE	256
#define MAXP	3

/* Magic numbers used for backtrace. */

#ifdef __i8086__
#	define ADDA(l) ((u16_t) (l) == 0xC481)
#	ifdef __i80386__
typedef int i32_t;
#		define ADDA_CNT(l) ((i32_t) (l))
#	else
typedef int i16_t;
#		define ADDA_CNT(l) ((i16_t) (l))
#	endif
#	define ADDQ(l) ((u16_t) (l) == 0xC483)
#	define ADDQ_CNT(l) (((((l) >> 16) + 128) & 0x000000FF) - 128)
#	define BREAK(l) (0x000000CC | ((l) & 0xFFFFFF00))
#	define BREAKPOINT_ADVANCE 1
#	define INCSP2(l) ((u16_t) (l) == 0x4444)
#	define POPBX2(l) ((u16_t) (l) == 0x5B5B)
#	define POPBX(l)  ( (l & 0xFF) == 0x5B)
#endif

#ifdef __mc68000__
#	define ADDA(l) ((int)((l) >> 16) == 0xDFFC)
#	define ADDA_CNT(l) (l)
#	define ADDQ(l) ((((l) >> 16) & 0xF13F) == 0x500F)
#	define ADDQ_CNT(l) (((((l) >> 25) - 1) & 7) + 1)
#	define BREAK(l) (0xA0000000 | ((l) & 0xFFFF))
#	define BREAKPOINT_ADVANCE 0
#	define LEA(l) (((l) >> 16) == 0x4FEF)
#	define LEA_DISP(l) ((long)(int)(l))
#	define LINK(l) (((l) >> 16) == 0x4E56)
#	define RTS(l) (((l) >> 16) == 0x4E75)
#endif

PRIVATE long lastexp = -1L;	/* last expression and segment */
PRIVATE int lastseg = T;
PRIVATE struct sgttyb in_gttyb, sv_gttyb; /* initial and saved */
PRIVATE struct tchars in_chars, sv_chars;
PRIVATE char cbuf[MAXLINE];
PRIVATE char sbuf[MAXLINE];
PRIVATE char obuf[MAXLINE];
PRIVATE int rw = 0;
PRIVATE int ei = 1;
#define DI { int sei = ei; ei = 0;
#define EI ei = sei; }

struct b_pnt
{
	struct b_pnt *nxt, *prv;
	long addr;
	long oldval;
	char cmd[1];
} *b_head, *curpnt;

FORWARD void cleanup P_((void));
FORWARD int dowait P_((void));
FORWARD void update P_((int dotty));
FORWARD void findbpnt P_((int verbose));
FORWARD int exebpnt P_((int restart));
FORWARD void tstart P_((int req, int verbose, int val, int cnt));
FORWARD void catch P_((int sig));
FORWARD int run P_((char *name, char *argstr, int tflg));
FORWARD void backtrace P_((int all));
FORWARD void freepnt P_((struct b_pnt *pnt));
FORWARD void breakpt P_((long addr, char *local_cmd));
FORWARD void resetbpnt P_((void));
FORWARD void modify P_((long addr, int cnt, int verbose, int size));
FORWARD void display P_((long addr));
FORWARD void fill P_((long addr, int req));
FORWARD void dorun P_((char *local_cmd));
FORWARD void prthelp P_((void));
FORWARD void command P_((void));

PRIVATE void cleanup()
{
	rw = curpid = 0;
	core_init((char *) NULL, prog, 0, 0);
	curpnt = NULL;
}

PRIVATE int dowait()
{
	int stat;

	DI;
	while (wait(&stat) != curpid) {};
	EI;
	lastexp = -1;
	if (WIFEXITED(stat))
	{
		if (WEXITSTATUS(stat) != 0177)
			printf("child exited with status %d\n",
			       WEXITSTATUS(stat));
		cleanup();
		return 0;
	}
	if (WIFSIGNALED(stat))
	{
		printf("child terminated by signal %d\n", WTERMSIG(stat));
#define WIFCORED(stat) (stat & 0x80)
		if (WIFCORED(stat)) printf("(core dumped)\n");
		cleanup();
		return 0;
	}
	return cursig = WSTOPSIG(stat);
}

PRIVATE void update(dotty)
	int dotty;
{
	int i;

	if (dotty)
	{
		ioctl(0, TIOCGETP, &sv_gttyb);
		ioctl(0, TIOCGETC, (struct sgttyb *) &sv_chars);
		ioctl(0, TIOCSETP, &in_gttyb);
		ioctl(0, TIOCSETC, (struct sgttyb *) &in_chars);
	}
	for (i = 0; i < sizeof lbuf / sizeof lbuf[0] - 1; i++)
	{
		lbuf[i] = myptrace(T_GETUSER, curpid,
				   (long)(i*sizeof(long)), 0L);
	}
	saddr = ksize_to_physb(prc->p_map[T].mem_vir);
	daddr = ksize_to_physb(prc->p_map[D].mem_vir);
	taddr = ksize_to_physb(prc->p_map[S].mem_vir);
	eaddr = ksize_to_physb(prc->p_map[D].mem_vir + prc->p_map[D].mem_len);
}

PRIVATE void findbpnt(verbose)
	int verbose;
{
	for (curpnt = b_head; curpnt; curpnt = curpnt->nxt)
	{
		if (curpnt->addr + saddr == PC_MEMBER(prc) - BREAKPOINT_ADVANCE)
		{
			myptrace(T_SETINS, curpid, curpnt->addr + saddr, curpnt->oldval);
			myptrace(T_SETUSER, curpid, PC_OFF, curpnt->addr + saddr);
			if (curpnt->cmd[0] != '\n')
				cmd = strcpy(cbuf, curpnt->cmd);
			else if (verbose)
				printf("Breakpoint hit.\n");
			return;
		}
	}
	if (verbose) printf("Unknown breakpoint hit.\n");
}

PUBLIC int disp_regs()
{
#ifdef MINIX_PC
	register struct stackframe_s *regptr = &prc->p_reg;
#else
	int i;
	register reg_t *regptr = (reg_t *) &prc->p_reg;
#endif

	if (curpid <= 0)
	{
		printf("No active process.\n");
		return 1;
	}

/* The MINIX_PC reg printing routines were originally written for sdump.c,
 * then duplicated (almost :-() for dmp.c, db.c, here, and may later be
 * duplicated for dcore.c and postmort.c.  They should be in a library.
 */
#if defined(MINIX_PC) && defined(__i80386__)
	printf("eax=%8lX   ebx=%8lX   ecx=%8lX   edx=%8lX\n",
	       regptr->retreg, regptr->bx, regptr->cx, regptr->dx);
	printf("esi=%8lX   edi=%8lX   ebp=%8lX   esp=%8lX   eip=%8lX\n",
	       regptr->si, regptr->di, regptr->bp, regptr->sp, regptr->pc);
	printf(" ds=%8lX    es=%8lX                   ss=%8lX    cs=%8lX\n",
	       regptr->ds, regptr->es, regptr->ss, regptr->cs);
	printf(" fs=%8lX    gs=%8lX                                  ef=%8lX\n",
	       regptr->fs, regptr->gs, regptr->psw);
#endif

#if defined(MINIX_PC) && !defined(__i80386__)
	printf(
"ax=%4X  bx=%4X  cx=%4X  dx=%4X  si=%4X  di=%4X  bp=%4X  sp=%4X  ip=%4X\n",
	       regptr->retreg, regptr->bx, regptr->cx, regptr->dx,
	       regptr->si, regptr->di, regptr->bp, regptr->sp, regptr->pc);
	printf(
" f=%4X                             ds=%4X  es=%4X           ss=%4X  cs=%4X\n",
	       regptr->psw, regptr->ds, regptr->es, regptr->ss, regptr->cs);
#endif

#ifdef MINIX_ST
	printf("\npc=%lx psw=%x\n\n",(long)PC_MEMBER(prc), PSW_MEMBER(prc));
	printf(
"      0        1        2        3        4        5        6        7\nD");
	for (i = 0; i < 8; i++) printf(" %08lx", *regptr++);
	printf("\nA");
	for (; i < NR_REGS; i++) printf(" %08lx", *regptr++);
	printf(" %08lx\n\n", (long)SP_MEMBER(prc));
#endif

	return 0;
}

PUBLIC char *skip(s)
	register char *s;
{
	while (isspace(*s)) ++s;
	return *s ? s : s-1;
}

PRIVATE int exebpnt(restart)
	int restart;
{
	myptrace(T_STEP, curpid, 0L, (long)restart);
	if (dowait() == 0) return 1;
	myptrace(T_SETINS, curpid, curpnt->addr + saddr, BREAK(curpnt->oldval));
	curpnt = 0;
	return 0;
}

PRIVATE void tstart(req, verbose, val, cnt)
	int req, verbose, val, cnt;
{
	if (curpid == 0 || curpid == 1)
	{
		if (req == T_EXIT && curpid)
		{
			myptrace(T_EXIT, curpid, 0L, (long)val);
			curpid = 0;
		}
		if (verbose) printf("No active process.\n");
		return;
	}
	if (req == T_EXIT)
	{
		myptrace(T_EXIT, curpid, 0L, (long)val);
		dowait();
		return;
	}
	if (cnt==0) cnt = 1;
	ioctl(0, TIOCSETP, &sv_gttyb);
	ioctl(0, TIOCSETC, (struct sgttyb *) &sv_chars);
	do
	{
		if (curpnt)
		{
			if (exebpnt(val)) return;
			if (req == T_RESUME) cnt++;
			val = 0;
		}
		else
		{
			myptrace(req, curpid, 0L, (long)val);
			if (dowait() == 0) return;
			val = 0;
			switch (cursig)
			{
			case	SIGEMT: /* breakpoint */
				update(0);
				findbpnt(cnt <= 1);
				break;
			case	SIGTRAP: /* trace trap? */
				if (req == T_STEP) break;
			default: /* signal */
				val = cursig;
				break;
			}
		}
	}
	while (--cnt > 0);
	update(1);
	dasm((long)PC_MEMBER(prc), 1, 1);
}

PRIVATE void catch(sig)
	int sig;
{
	signal(sig, catch);
	if (sig == SIGINT || sig == SIGQUIT)
	{
		if (ei)
			longjmp(mainlp, 1);
		else
			return;
	}
	tstart(T_EXIT, 0, sig, 0);
	exit(0);
}

PRIVATE int run(name, argstr, tflg)
	char *name, *argstr;
	int tflg;
{
	int procid;

	if( (procid = fork()) == 0 )
	{
		char *argv[MAXARG], *inf = 0, *outf = 0;
		int argc;

		if (tflg && myptrace(T_OK, 0, 0L, 0L) < 0)
		{
			perror("enable trace");
			exit(127);
		}
		argv[0] = name;
		for (argc = 1; ; )
		{
			argstr = skip(argstr);
			if (*argstr == '\n' || *argstr == ';')
			{
				argv[argc] = 0;
				if (inf) freopen(inf, "r", stdin);
				if (outf) freopen(outf, "w", stdout);
				if (tflg)
				{
					execv(name, argv);
					perror("execv");
				}
				else
				{
					execvp(name, argv);
					perror("execvp");
				}
			        exit( 127 );
			}
			if (*argstr == '<') inf = argstr+1;
			else if (*argstr == '>') outf = argstr+1;
			else if (argc == MAXARG)
			{
				printf("Too many arguments.\n");
				exit(127);
			}
			else argv[argc++] = argstr;
			while (!isspace(*argstr)) argstr++;
			if (*argstr == '\n') argstr[1] = '\n', argstr[2] = 0;
			*argstr++ = 0;
		}
	}
	if (procid < 0)
	{
		printf("Fork failed.\n");
	}
	return procid;
}

PRIVATE void backtrace(all)
	int all;
{
	long pc, bp, off, val, obp, opc;
	int flag = 0;
	if (curpid <= 0)
	{
		printf("No process.\n");
		return;
	}
	pc = PC_MEMBER(prc);
#ifdef __mc68000__
	off = myptrace(T_GETINS, curpid, pc, 0L);
	if (LINK(off) || RTS(off))
	{
		bp = SP_MEMBER(prc) - ADDRSIZE;
		flag = 1;
	}
	else
#endif
		bp = BP_MEMBER(prc);

	if (bp == 0)
	{
		printf("No active frame.\n");
		return;
	}
	errno = 0;
	do
	{
		opc = pc;
		pc = (myptrace(T_GETDATA, curpid, bp+ADDRSIZE, 0L)
		      >> SHIFT(ADDRSIZE)) & MASK(ADDRSIZE);
		errno = 0;	/* XXX */
		off = myptrace(T_GETINS, curpid, pc, 0L);
		if (errno != 0) return;
		symbolic(opc, '(');
		obp = bp;
		bp += 2*ADDRSIZE;
		if (ADDQ(off))
			off = ADDQ_CNT(off) + bp;
#ifdef __mc68000__
		else if (LEA(off))
			off = LEA_DISP(off) + bp;
#endif
		else if (ADDA(off))
			off = ADDA_CNT(myptrace(T_GETINS, curpid, pc+2, 0L)) + bp;
#ifdef __i8086__
		else if (INCSP2(off))
			off = 2 + bp;
		else if (POPBX2(off))
			off = 2 + bp;
		else if (POPBX(off))
			off = 1 + bp;
#endif
		else
			goto skiplp;

		for (;;)
		{
			if (errno) return;
			val = ((myptrace(T_GETDATA, curpid, bp, 0L)
			       >> SHIFT(INTSIZE))) & MASK(INTSIZE);
			printf("0x%0*x", 2 * (int)sizeof(int), (int)val);
			bp += INTSIZE;
			if (bp >= off) break;
			putc(',', stdout);
		}

		skiplp:

		fputs(")\n", stdout);
		if (flag)
		{
			bp = BP_MEMBER(prc);
			flag = 0;
		}
		else
			bp = myptrace(T_GETDATA, curpid, obp, 0L);
	}
	while (all && (reg_t) bp );
}

PRIVATE void freepnt(pnt)
	struct b_pnt *pnt;
{
	if (pnt->prv) pnt->prv->nxt = pnt->nxt;
	else b_head = pnt->nxt;
	if (pnt->nxt) pnt->nxt->prv = pnt->prv;
	if (curpid > 0) myptrace(T_SETINS, curpid, pnt->addr + saddr, pnt->oldval);
	free(pnt);
	if (pnt == curpnt) curpnt = 0;
}

PRIVATE void breakpt(addr, local_cmd)
	long addr;
	char *local_cmd;
{
	struct b_pnt *new;

	for (new = b_head; new; new=new->nxt)
		if (new->addr + saddr == addr)
		{
			printf("Breakpoint already exists here.\n");
			return;
		}
	new = (struct b_pnt *) malloc((size_t) (sizeof(struct b_pnt)
						+ strlen(local_cmd)));
	if (new == 0)
	{
		printf("No room.\n");
		return;
	}
	new->nxt = b_head;
	new->prv = 0;
	if (b_head) b_head->prv = new;
	b_head = new;
	new->addr = addr - saddr;
	strcpy(new->cmd, local_cmd);
	if (curpid > 1)
	{
		new->oldval = myptrace(T_GETINS, curpid, addr, 0L);
		myptrace(T_SETINS, curpid, addr, BREAK(new->oldval));
		if (myptrace(T_GETINS, curpid, addr, 0L) != BREAK(new->oldval))
		{
			perror("Can't set breakpoint");
			freepnt(new);
		}
	}
}

PRIVATE void resetbpnt()
{
	struct b_pnt *bpnt;

	for (bpnt = b_head; bpnt; bpnt = bpnt->nxt)
	{
		bpnt->oldval = myptrace(T_GETINS, curpid, bpnt->addr + saddr, 0L);
		myptrace(T_SETINS, curpid, bpnt->addr + saddr, BREAK(bpnt->oldval));
		if (myptrace(T_GETINS, curpid, bpnt->addr + saddr, 0L) != BREAK(bpnt->oldval))
			perror("Can't reset breakpoint");
	}
}

PRIVATE void modify(addr, cnt, verbose, size)
	long addr;
	int cnt, verbose, size;
{
	long curval, off;

	if (curpid <= 1)
	{
		printf("No active process.\n");
		return;
	}
	curval = (myptrace(T_GETDATA, curpid, addr, 0L) >> SHIFT(INTSIZE))
		 & MASK(size);
	ioctl(0, TIOCSETP, &sv_gttyb);
	ioctl(0, TIOCSETC, (struct sgttyb *) &sv_chars);
	do
	{
		if (cursig == SIGTRAP) cursig = 0;
		if (verbose)
		{
			off = get_reg(curpid, PC_OFF);
			dasm(off, 1, 0);
		}
		if (curpnt && exebpnt(cursig)) return;
		else
		{
			myptrace(T_STEP, curpid, addr, 0L);
			switch (dowait())
			{
			case	0: return;
			case	SIGEMT: update(0); findbpnt(0); break;
			}
		}
		if (curval != (myptrace(T_GETDATA, curpid, addr, 0L)
			       >> SHIFT(INTSIZE)) & MASK(size) )
		{
			printf("Modification detected\n");
			break;
		}
	}
	while (--cnt);
	update(1);
	dasm((long)PC_MEMBER(prc), 1, 1);
}

PRIVATE void display(addr)
	long addr;
{
	int count, size, out, shift, req;
	long val, msk;
	char fmt;

	if (curpid < 0)
	{
		printf("No active process\n");
		return;
	}
	
	req = (seg == T ? T_GETINS : T_GETDATA);
	count = strtol(cmd, &cmd, 0);
	if (count == 0) count = 1;
	cmd = skip(cmd);
	if (*cmd == 'i' || *cmd == 'I')
	{
		lastexp = dasm(addr, count, *cmd == 'i');
		return;
	}
	switch (*cmd++)
	{
	case	'b': size = sizeof(char); break;
	case	'h': size = sizeof(short); break;
	case	'l': size = sizeof(long); break;
	default	   : size = sizeof(int); --cmd; break;
	}
	switch (fmt = *cmd)
	{
	case	'c': count *= size; size = sizeof(char); break;
	case	's': addr = myptrace(req, curpid, addr, 0L); 
	case	'a': size = sizeof(char);
		     break;
	}
	out = 0;
	msk = MASK(size);
	shift = SHIFT(size);
	do
	{
		val = (myptrace(req, curpid, addr, 0L) >> shift) & msk;
		if (out == 0) printf("\n0x%0*lx: ", (int) (2 * ADDRSIZE),
				(addr >> SHIFT(ADDRSIZE)) & MASK(ADDRSIZE) );
		switch (fmt)
		{
		case	'c':
			printf(isprint((unsigned char) val)
			       ? "   %c " : "\\%03o ", (unsigned char) val);
			if (++out == 8) out = 0;
			break;
		case	'u':
			printf("%12lu ", val); if (++out == 4) out = 0; break;
		case	'x':
			printf("%*lx ", 2*size, val);
			if (++out == (size == 4 ? 4 : 8)) out = 0;
			break;
		case	'o':
			printf("%*lo ", 3*size, val);
			if (++out == (size == 4 ? 4 : 8)) out = 0;
			break;
		case	's':
		case	'a':
			if (val) fputc((char)val, stdout);
			else goto exitlp;
			if (++out == 64) out = 0;
			break;
		default:
		case	'd':
			printf("%12ld ", val); if (++out == 4) out = 0; break;
		}
		addr += size;
	}
	while (--count > 0 || fmt == 's' || fmt == 'a');
exitlp:
	lastexp = addr;
	fputc('\n', stdout);
}

PRIVATE void fill(addr, req)
	long addr;
	int req;
{
	char *cmdh;
	int count, size, shift;
	long val, msk, nval;

	if (!rw && curpid <= 1)
	{
		printf("No active process\n");
		return;
	}
	if (req == T_GETDATA && seg == T)
		req = T_GETINS;

	cmdh = cmd;
	count = strtol(cmd, &cmd, 0);

	switch (*cmd++)
	{
	case	'b': size = sizeof(char); break;
	case	'h': size = sizeof(short); break;
	case	'l': size = sizeof(long); break;
	default	   :
			if (req != T_GETUSER)
				size = sizeof(int);
			else
				size = sizeof(reg_t);
			--cmd; break;
	}
	cmd = skip(cmd);
	if (!*cmd || isspace(*cmd))
	{
		count = 1;
		size = sizeof(short);
		cmd = cmdh;
		cmd = skip(cmd);
	}
#ifdef BYTES_SWAPPED
	if (req == T_GETUSER)
	{
		shift = 0;
		msk = MASK(size);
	}
	else
	{
		shift = BITSIZE(LONGSIZE) - BITSIZE(size);
		msk = MASK(size) << shift;
	}
#else
	shift = SHIFT(size);
	msk = MASK(size);
#endif
	cmd = getexp(cmd, &nval, &seg);
	nval <<= shift;
	do
	{
		val = myptrace(req, curpid, addr, 0L) | (nval & msk);
		val &= (nval | ~msk);
		myptrace(req+3, curpid, addr, val);
		if (req == T_GETUSER)
			addr += sizeof(reg_t);
		else
			addr += size;
	}
	while (--count > 0);
	if (req == T_GETUSER)
		update(0);
}

PRIVATE void dorun(local_cmd)
	char *local_cmd;
{
	if (curpid = run(prog, local_cmd, 1))
	{
		if (dowait())
		{
			myptrace(T_SETUSER, curpid, BP_OFF, 0L);
			update(1);
			if (!b_head)
				printf("Process stopped.\n");
			else
			{
				resetbpnt();
				tstart(T_RESUME, 0, 0, 1);
			}
		}
	}
}

/* print short help message */
PRIVATE void prthelp()
{
	printf("\n");
	printf("! \t- Set data variable\n");
	printf("# \t- Set register\n");
	printf("Tt\t- Backtrace\n");
	printf("= \t- Print as symbol/number\n");
	printf("/ \t- Display\n");
	printf("Xx\t- Unasm\n");
	printf("Rr\t- Run\n");
	printf("Cc\t- Continue\n");
	printf("Ii\t- Single step\n");
	printf("Mm\t- Stop when modified\n");
	printf("k \t- Kill\n");
	printf("Bb\t- Display / Set Break-pt\n");
	printf("Dd\t- Delete Break-pts\n");
	printf("V \t- Print version number\n");
#ifdef  DEBUG
	printf("v \t- Toggle debug flag\n");
#endif
	printf("e \t- List symbols\n");
	printf("y \t- Print mapping\n");
	printf("? \t- Help\n");
	printf("q \t- Quit\n");
}

PRIVATE void command()
{
	char c, *p;
	int i;
	int size;
	long exp;
	struct b_pnt *bp;

	seg = S; /* don't restrict segment expressions are in */
	cmdstart = cmd = skip(cmd);
	cmd = getexp(cmd, &exp, &seg);
	if (cmd == cmdstart)	/* expression ? */
	{
		if (lastexp < 0) {
			seg = T;
			exp = PC_MEMBER(prc);
		} else {
			seg = lastseg;
			exp = lastexp;
		}
	}
	if (cmd[0] == '\n' && cmd[1] == '\0')
		strcpy(cmd, obuf);
	else
		strcpy(obuf, cmd);

	lastexp = exp;	/* save last expression */
	lastseg = seg;

	/* check for legal commands */
	switch (c = *cmd++) {
	case '!':
	case 'T':
	case 't':
	case '/':
	case 'x':
	case 'X':
	case 'q':
	case '=':
	case ';':
	case 'V':
	case 'v':
	case 'e':
	case 'y':
	case '?':
	case 'r':
	case 'R':
	case 'k':
	case 'b':
	case 'B':
	case 'd':
	case 'D':	break;

	case '#':
			if (curpid <= 1) {
				printf("No active process\n");
				return;
			}
			break;

	default:
			if (curpid <= 1 ) {
				dorun("\n");
				return;
			}
			break;
	}

	switch (c)
	{
	case '!': /* escape to shell OR set variable to value */
		if (cmd == cmdstart+1)
		{
			DI;
			cmd = skip(cmd);
			if (*cmd == '\n' || *cmd == ';')
			{
				i = run("/bin/sh", "\n", 0);
			}
			else
			{
				for (p=cmd+1; *p && !isspace(*p); p++) {};
				*p++ = 0;
				i = run(cmd, *p ? p : "\n", 0);
			}
			if (i > 0) while (wait(&seg) != i) {};
			EI;
			break;
		}
		if (seg == T && (rw || curpid > 1))
			printf("Warning - modifying text\n");
		fill(exp, T_GETDATA);
		break;
	case 'T': /* top line of backtrace */
		backtrace(0);
		break;
	case 't': /* back trace */
		backtrace(1);
		break;
	case '/': /* print variable value */
		display(exp);
		break;
	case 'x': /* print registers and instruction */
		if (disp_regs()) break;
		/*FALLTHROUGH*/
	case 'X': /* print instruction */
		seg = T;
		exp = PC_MEMBER(prc);
		i = strtol(cmd, &cmd, 0);
		if (curpid > 0)
			dasm(exp, i ? i : 1, 1);
		else
			printf("No active process.\n");
		break;
	case 'R': /* run program with no args */
	case 'r': /* run program with args (possibly defaulted) */
		DI;
		tstart(T_EXIT, 0, 0, 0);
		if (c == 'r')
		{
			cmd = skip(cmd);
			if (*cmd == '\n' || *cmd == ';') cmd = sbuf;
			else strcpy(sbuf, cmd);
		}
		else
		{
			cmd = "\n";
		}
		dorun(cmd);
		EI;
		break;
	case 'c': /* continue program - ignore signal */
		cursig = 0;
	case 'C': /* continue program - handle signal */
		DI;
		i = 0;
		if (seg == T && curpnt == 0 && cmd != cmdstart+1)
		{
			breakpt(exp, "\n");
			curpnt = b_head;
			myptrace(T_SETINS, curpid, curpnt->addr + saddr, curpnt->oldval);
			i = 1;
		}
		tstart(T_RESUME, 1, cursig, (int)strtol(cmd, &cmd, 0));
		if (i) /* remove temporary bp */
		{
			freepnt(b_head);
		}
		if (cursig == SIGEMT) return;
		if (curpid > 1) printf("Process stopped by signal %d\n", cursig);
		EI;
		break;
	case 'i': /* single step - ignore signal */
		tstart(T_STEP, 1, 0, (int)strtol(cmd, &cmd, 0));
		break;
	case 'I': /* single step - handle signal */
		tstart(T_STEP, 1, cursig, (int)strtol(cmd, &cmd, 0));
		break;
	case 'm': /* single step until location modified */
	case 'M': /* single step until location modified - verbose */
		DI;
		cmd = skip(cmd);
		switch (*cmd++)
		{
		case	'b': size = sizeof(char); break;
		case	'h': size = sizeof(short); break;
		case	'l': size = sizeof(long); break;
		default	   : size = sizeof(long); --cmd; break;
		}
		modify(exp, (int)strtol(cmd, &cmd, 0), c == 'M', size);
		EI;
		break;
	case 'k': /* kill current program */
		tstart(T_EXIT, 1, 0, 0);
		break;
	case 'b': /* set a breakpoint at the given line */
#if 0
		if (seg != T)
		{
			printf("Address not in text space.\n");
			return;
		}
#endif
		breakpt(exp, skip(cmd));
		cmd = "\n";
		return;
	case 'B': /* print list of currently active breakpoints */
		for (i = 1, bp = b_head; bp; bp=bp->nxt, i++)
		{
			printf("%2d: ",i);
			symbolic((long) bp->addr + saddr,'\t');
			printf("(0x%lx)\t- %s", bp->addr + saddr, bp->cmd);
		}
		break;
	case 'd': /* delete breakpoint */
		if (seg == T)
		{
			for (bp = b_head; bp && bp->addr + saddr != exp; bp=bp->nxt) {};
			if (bp) { freepnt(bp); break; }
		}
		printf("No such breakpoint.\n");
		break;
	case 'D': /* delete all breakpoints */
		while(b_head) freepnt(b_head);
		break;
	case 'q': /* quit */
		tstart(T_EXIT, 0, 0, 0);
		exit(0);
	case '=':
		if (isdigit(*cmdstart)) symbolic(exp, '\n');
		else printf("0x%lx\n", exp);
	case ';':
		--cmd;
		break;
	case 'V': printf("MDB version 1.0\n"); break;
#ifdef	DEBUG
	case 'v': /* toggle debug */
		debug = !debug;
		if (debug) printf("Debug flag ON\n");
		break;
#endif
	case 'e': /* list symbols */
		listsym(cmd); break;
	case 'y': /* print mapping */
		prtmap(); break;
	case '?': /* print help    */
		prthelp(); break;
	case '#': /* set register  */
		cmdstart = cmd;
		cmd = skip(cmd+2);
		if (*cmd == '!')
		{
			cmd++;
			fill(reg_addr(cmdstart), T_GETUSER);
			break;
		}
		/*FALLTHROUGH*/
	default:
		printf("Unknown command.\n");
		break;
	}
	while (*cmd != '\n' && *cmd != ';') ++cmd;
	if (*cmd == ';') cmd = skip(cmd+1);
}

/* extern int getopt P_((int argc, char **argv, char *optstring)); */
/* Use prototype from stdlib (gcc: unistd.h) */

extern char *optarg;
extern int optind;

PUBLIC int main(argc, argv)
	int argc;
	char *argv[];
{
	int i, c, flat = 0;

	while ((c = getopt(argc, argv, "wf?")) != EOF) {
		switch (c) {
			case 'w':   rw   = 1;    break;
			case 'f':   flat = 1;    break;

			case '?':
				fprintf(stderr, "Usage: %s [-w] [-f] [file [corefile]]\n",
					argv[0]);
			exit(1);
		}
	}
	argv += optind - 1;
	argc -= optind - 1;

	prc = (struct proc *) lbuf;
	ioctl(0, TIOCGETP, &in_gttyb);
	ioctl(0, TIOCGETC, (struct sgttyb *) &in_chars);

	for (i = 1; i < _NSIG; i++) signal(i, catch);
	strcpy(sbuf, "\n");
	prog = argc > 1 ? argv[1] : "a.out";
	core = argc > 2 ? argv[2] : "core";

	getsyms(prog);

	if ((curpid = core_init(core, prog, rw, flat)) >= 0) {
		update(0);
	}

	if (setjmp(mainlp))
	{
		ei = 1;
		printf("\n");
	}

	while ((printf("* "), fflush(stdout),
			fgets(cbuf, sizeof(cbuf), stdin)) != NULL)
	{
	start_loop:
		if (strlen(cbuf) == sizeof(cbuf)-1)
		{
			printf("Command line too long.\n");
			continue;
		}
		if (cbuf[strlen(cbuf)-2] == '\\')
		{
			printf("> "), fflush(stdout);
			fgets(cbuf+strlen(cbuf)-2, (int)sizeof(cbuf)-strlen(cbuf)-2,
				stdin);
			goto start_loop;
		}
		cmd = cbuf;
		command();
		while (*cmd != '\n') command();
	}
	tstart(T_EXIT, 0, 0, 0);
	return 0;
}
