#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include <sgtty.h>
#include <screen.h>
#include <st_screen.h>
#include "proc.h"
#include "staddr.h"
#include "stvideo.h"

/* hardware independend prototypes */
_PROTOTYPE( int do_screen, (message *mess) );
_PROTOTYPE( PRIVATE int scr_open, (message *mp) );
_PROTOTYPE( PRIVATE int scr_close, (message *mp) );
_PROTOTYPE( PRIVATE int scr_ioctl, (message *mp) );

/* hardware specific prototypes */
_PROTOTYPE( PRIVATE struct rgb_value get_color_map_entry, (int index) );
_PROTOTYPE( PRIVATE short color_map_entry, (struct rgb_value rgb_value) );

/* size of color lookup table buffer */
#define MAX_CLUT_SIZE		(256)

/* NR_SCREENS is the number of screen devices actually implemented */
#define DIM(a)		(sizeof(a)/sizeof(a[0]))
#define NR_SCREENS	DIM(video_mode)

/* VIDEO_RES defines the hardware resolution register */
#if (ATARI_TYPE == TT)
#define VIDEO_RES	VIDEO->vd_tt_res
#define VIDEO_RGB	VIDEO->vd_tt_rgb
#else
#define VIDEO_RES	VIDEO->vd_st_res
#define VIDEO_RGB	VIDEO->vd_st_rgb
#endif

/* cur_mode holds the number of the current video mode. -1 = TEXT */
PRIVATE struct rgb_value color_map[MAX_CLUT_SIZE];	/* temp buffer for clut */
PRIVATE short previous_video_mode;	/* video mode before opening device */
PRIVATE int  open_count;		/* # times the device is opened */

#ifndef DEV_OPEN
/* Kludge for MINIX 1.5.10 */
#define DEV_OPEN	TTY_OPEN
#define DEV_CLOSE	TTY_CLOSE
#define DEV_IOCTL	TTY_IOCTL
#endif

/*===========================================================================*
 *				do_screen				     * 
 *===========================================================================*/
PUBLIC int do_screen(mp)
register message *mp;
{
  int r;

  if (mp->DEVICE < SCREEN_DEV || mp->DEVICE >= SCREEN_DEV+NR_SCREENS)
	return ENXIO;
  switch(mp->m_type) {
	case DEV_OPEN:	r = scr_open(mp);	break;
	case DEV_CLOSE:	r = scr_close(mp);	break;
	case DEV_IOCTL:	r = scr_ioctl(mp);	break;
	default:	r = EINVAL;		break;
  }
  return(r);
}

/*===========================================================================*
 *				scr_open				     *
 *===========================================================================*/
PRIVATE int scr_open(mp)
register message *mp;
{
  static int current_minor;	/* used for rejecting other open-s */
  int  index;
  /* if the device is not in use open it */
  if (open_count == 0)
  {
	index = mp->DEVICE - SCREEN_DEV;
	current_minor = mp->DEVICE;
	previous_video_mode = VIDEO_RES;
	VIDEO_RES = video_mode[index].mode;
	video_mode[index].attr.addr = (((long)(VIDEO->vd_ramh & 0xFF) << 16) |
				       ((long)(VIDEO->vd_ramm & 0xFF) << 8) );
  }
  else
  {
	/* if the device is in use but through another minor, reject */
	if (current_minor != mp->DEVICE) return(EBUSY);
  }
  open_count++;
  return(OK);
}

/*===========================================================================*
 *				scr_close				     *
 *===========================================================================*/
PRIVATE int scr_close(mp)
register message *mp;
{
  /* release the device, and if it is the last open, restore the video mode */
  if (--open_count == 0)
  {
	VIDEO_RES = previous_video_mode;
  }
  return(OK);
}

/*===========================================================================*
 *				scr_ioctl				     *
 *===========================================================================*/
PRIVATE int scr_ioctl(mp)
register message *mp;
{
/* Handle ioctl commands for screen devices */
  struct clut_entry entry;
  vir_bytes nbytes, addr;
  phys_bytes user_phys, my_phys;
  int from_user;			/* flag for phys_copy */
  int index;
  int i;

  index = mp->DEVICE - SCREEN_DEV;
/* First gather address and number of bytes to copy. */
  switch(mp->TTY_REQUEST) {
	case SCR_ATTR:
		addr = (vir_bytes) &(video_mode[index].attr);
		nbytes = sizeof(struct scr_attr);
		from_user = 0;
		break;
	case SCR_GETCOLMAP:
		for (i = 0; i < video_mode[index].attr.clut_size; i++)
		{
		    color_map[i] = get_color_map_entry(i);
		}
		addr = (vir_bytes) color_map;
		nbytes = video_mode[index].attr.clut_size *
			 sizeof(struct clut_entry);
		from_user = 0;
		break;
	case SCR_SETCOLMAP:
		addr = (vir_bytes) color_map;
		nbytes = video_mode[index].attr.clut_size *
			 sizeof(struct clut_entry);
		from_user = 1;
		break;
	case SCR_GETCOL:
	case SCR_SETCOL:
		addr = (vir_bytes) &entry;
		nbytes = sizeof(entry);
		from_user = 1;
		break;
	default:
		return(EINVAL);
  }
  user_phys = numap(mp->PROC_NR, (vir_bytes) mp->ADDRESS, nbytes);
  if (user_phys == 0)
	return(E_BAD_ADDR);
  my_phys = umap(proc_ptr, D, addr, nbytes);
  if (from_user)
	phys_copy(user_phys, my_phys, nbytes);
  else
	phys_copy(my_phys, user_phys, nbytes);
/* now perform the required setting (if one is required) */
  switch (mp->TTY_REQUEST) {
	case SCR_GETCOL:
		if (entry.index > video_mode[index].attr.clut_size)
			return(EINVAL);
		entry.rgb = get_color_map_entry(entry.index);
		phys_copy(my_phys, user_phys, nbytes);
		break;
	case SCR_SETCOL:
		if (entry.index > video_mode[index].attr.clut_size)
			return(EINVAL);
		VIDEO_RGB[entry.index] = color_map_entry(entry.rgb);
		break;
	case SCR_SETCOLMAP:
		for (i = 0; i < video_mode[index].attr.clut_size; i++)
		{
		    VIDEO_RGB[i] = color_map_entry(color_map[i]);
		}
		break;
  }
  return(OK);
}

PRIVATE struct rgb_value get_color_map_entry(index)
int index;
{
    struct rgb_value rgb_value;
    int tmp;
    
    tmp = VIDEO_RGB[index];
#if (ATARI_TYPE == STe)
    if (tmp & 0x8) rgb_value.b = (tmp & 7) << 1 + 1;
    else rgb_value.b = (tmp & 7) << 1;
    tmp = tmp >> 4;
    if (tmp & 0x8) rgb_value.g = (tmp & 7) << 1 + 1;
    else rgb_value.g = (tmp & 7) << 1;
    tmp = tmp >> 4;
    if (tmp & 0x8) rgb_value.r = (tmp & 7) << 1 + 1;
    else rgb_value.r = (tmp & 7) << 1;
#else
    rgb_value.r = (tmp >> 8) & 0xf;
    rgb_value.g = (tmp >> 4) & 0xf;
    rgb_value.b = (tmp     ) & 0xf;
#endif
    return(rgb_value);
}

PRIVATE short color_map_entry(rgb_value)
struct rgb_value rgb_value;
{
#if (ATARI_TYPE == STe)
    short result = 0;
    
    if (rgb_value.r) result = rgb_value.r >> 1 + 0x8;
    else result = rgb_value.r >> 1 + 0x8;
    result << 4;
    if (rgb_value.g) result = result + rgb_value.g >> 1 + 0x8;
    else result = result + rgb_value.g >> 1 + 0x8;
    result << 4;
    if (rgb_value.b) result = result + rgb_value.b >> 1 + 0x8;
    else result = result + rgb_value.b >> 1 + 0x8;
    return(result);
#else
    return((rgb_value.r << 8) + (rgb_value.g << 4) + rgb_value.b);
#endif
}
