/*
 * The video driver for the Atari ST
 */

/* This driver does not deal with multiple consoles and parameter passing
   through tp. Also the tty struct fields row and column are not maintained.

   Multiple console implemented. Also some code cleanups and speedups have
   been done. Cursor control is now local to this file. There are other parts
   of this source, such as the escape sequence handling and the character
   painting functions, which need a cleanup. F.e. the pc-minix specific
   fields should be deleted from tty_struct, and a pointer to the virterm
   structure should be inserted there.			K.-U. Bloem, --kub

   Added/corrected VT100 compatibility - Now looks _almost_ exactly like a
   DEC VT100 to an application.
   Added scroll regions, tab set/clear, underline, disable/enable cursor.
   Added escape code 'esc[24~' to set 24-line mode.
     "     "     "   'esc[25~' to set 25-line mode.
   Still no 132 column, or graphics characters, or device status reports,
   or double height/width lines, or vt52 mode.
   Otherwise, it passes all tests in vttest.c
   				jdoss@jmdst.lonestar.org - December, 1991

   Added BOLD code, and re-worked paint() function back to one generic
   function, rather than resolution-specific functions.
      				jdoss@jmdst.lonestar.org - February, 1992
   Support more resolutions. Made it a lot easier to add new resolutions.
   added reduced intensity, PIXEL WONDER code (untested)
   clean up; chosen for simplicity at the expense of a little code size
   for monochrome screens.
   Some performance gain can be achieved by adding a paintfunc field
   in the virterm struct and making different paint functions for 
   bold, light, normal, different font sizes etc. Then when the
   attributes change, the paintfunc field can be loaded with the
   appropriate paint function. Instead of paint then v->paintfunc()
   can be called. For the time being I decided that this was not worth
   the additional code.
				      Frans Meulenbroeks - November, 1992

*/

/* TODO:
	- perhaps: assembly paint func
	- perhaps: code 8 tab stops in a byte (worth the effort?) 
	- test TT hi res 
	- if color screen light/bold attributes by using different gray colors
*/

#include "kernel.h"
#if (MACHINE == ATARI)
#include <sgtty.h>
#include <minix/com.h>

#include "staddr.h"
#include "stvideo.h"
#include "stsound.h"

#include "tty.h"

/* default screen colors - change these to your taste */
#define DEFAULT_BKGRND	RGB_WHITE
#define DEFAULT_FORGRND RGB_BLACK

/* Constants relating to the video RAM and fonts */

#if (ATARI_TYPE == TT)
#define	VIDEORAM	320*480L /* size of video ram */
#else
#define	VIDEORAM	32000L	/* size of video ram */
#endif
#define VIDEOALIGNMENT	256	/* alignment of video ram */
#define NCOL		80	/* characters on a row */

/* bit values for v->attr */
#define REVERSE		1
#define UNDERLINE	2
#define BOLD		4
#define LIGHT		8
#define DISPLAY_ATTRIBS (REVERSE|UNDERLINE|BOLD|LIGHT)

/* structure to describe the characteristics of the various video modes */
struct vdu_params {
	unsigned char  nr_planes;
	unsigned short nr_bytes_per_line;
	unsigned short nr_video_lines;
};

PRIVATE struct vdu_params vdu_params[] = {
	{ 4,  320/8, 200 },	/* ST_RES_LOW  */
	{ 2,  640/8, 200 },	/* ST_RES_MED  */
	{ 1,  640/8, 400 },	/* ST_RES_HIGH */
	{ 0,      0,   0 },	/* unused      */
	{ 4,  640/8, 480 },	/* TT_RES_MED  */
	{ 0,      0,   0 },	/* unused      */
	{ 1, 1280/8, 960 },	/* TT_RES_HIGH */
	{ 8,  320/8, 480 }	/* TT_RES_LOW  */
};
/* Structure to keep data about virtual consoles in. Each virtual console
 * has its own screen ram, color table, keyboard setting etc. Exchanging the
 * currently displayed virtual console just requires a reprogramming of the
 * video controller with the new data from virterm, as well as setting some
 * variables for the keyboard driver.
 */
PRIVATE struct virterm {
		/* vdu hardware controller data */
	unsigned short vrgb[256];/* color table entry */
	unsigned short vres;	/* screen resolution */
	unsigned char  nr_planes;	/* 1 if mono, > 1 if color */
	char	*vram;		/* base of video ram */
		/* font and cursor data */
	int	linc;		/* video lines in char (font height) */
	int	bytc;		/* bytes in video line */
	int	nrow;		/* char rows on screen */
	int	bytr;		/* bytes in row of chars */
	char	*curs;		/* cursor position in video RAM */
	int	curmode;	/* cursor mode, 1 = cursor on */
	int	cursenable;	/* cursor enabled, 0 = cursor invisible */
		/* terminal output state */
	unsigned char	attr;	/* current attribute byte */
	char	tabs[NCOL];	/* current tab stops */
	int	topscroll;	/* top line of scroll region */
	int	botscroll;	/* bottom line of scroll region */
	int	ccol;		/* current char column */
	int	crow;		/* current char row */
	int	wrap;		/* state of auto-wrap */
	int	origmode;	/* origin mode */
	unsigned char savattr;	/* saved attribute byte */
	int	savccol;	/* saved char column */
	int	savcrow;	/* saved char row */
	int	savorigmode;	/* saved origin mode */
	char	vbuf[30];	/* partial escape sequence */
	char	*next;		/* next char in vbuf[] */
		/* terminal input state */
	int	vkeypad;	/* keyboard keypad state */
	int	vapmode;	/* keyboard cursor state */
} virterm[NR_CONS];

/* If there is only one plane (monochrome) then all bytes are adjacent in
   memory. However, if there are more planes then in memory a word (2 bytes)
   from a plane is stored, followed by a word for every other plane. This 
   means that when advancing a byte within a plane, for even bytes we have to
   increment the position in memory by one, while for odd bytes we have to
   increment the position with (nr_planes - 1) * 2 + 1.
   The +1 increments while the (nr_planes - 1) * 2 skips the other planes.
   To avoid that the calculation (nr_planes - 1) * 2 + 1 occurs often it
   is cached into the variable increment_to_next_word. 
*/
int increment_to_next_word;

char statline[] =
"- - - - - - - - - - - - - - -  24 line mode  - - - - - - - - - - - - - - - - - -";

/* This is a pointer to the currently displayed virtual console device. The
 * keyboard driver needs to know to which tty it must deliver keys from the
 * keyboard. Moreover the kernel sometimes needs to output some text. This
 * should go to the current console device, since things may block after
 * text output.
 */

PUBLIC struct tty_struct *cur_cons = &tty_struct[0];

/* the next four variables are for synchronising the setting of the
   video registers with the vertical blanking. Not synchronising 
   gives problems on Mega STe's */
unsigned int vid_res;
unsigned short vid_rgb [256];
char *vid_ram;
int progvdu = 0; /* make 1 to setup video hardware at next vbl */

unsigned short resolution; /* screen resolution; ST_RES_LOW...TT_RES_HIGH */

/* forward declarations for internal procedures */

/* general char output driver and escape sequence handling */
FORWARD _PROTOTYPE( void vdu_char, (struct tty_struct *tp,
			struct virterm *v, int c)			);
FORWARD _PROTOTYPE( void vductrl, (struct tty_struct *tp,
			struct virterm *v, int c)			);
FORWARD _PROTOTYPE( void vduesc, (struct tty_struct *tp,
			struct virterm *v, int c)			);
FORWARD _PROTOTYPE( void vduansi, (struct tty_struct *tp,
			struct virterm *v, int c)			);
FORWARD _PROTOTYPE( int vduparam, (struct virterm *v)			);
/* character printing and cursor management */
FORWARD _PROTOTYPE( void vducursor, (struct virterm *v, int onoff)	);
FORWARD _PROTOTYPE( void paint, (struct virterm *v, int c)		);
FORWARD _PROTOTYPE( void beep, (void)					);
FORWARD _PROTOTYPE( void moveto, (struct virterm *v, int r, int c)	);
/* screen modifications apart from char printing */
FORWARD _PROTOTYPE( void vscroll, (struct virterm *v,
			int up, int rl, int ru, int n)			);
FORWARD _PROTOTYPE( void hscroll, (struct virterm *v,
			int rt, int r, int cl, int cu, int n)		);
FORWARD _PROTOTYPE( void clrarea, (struct virterm *v,
			int rl, int cl, int ru, int cu)			);
FORWARD _PROTOTYPE( void clrlines, (struct virterm *v, int r, int n)	);
FORWARD _PROTOTYPE( void clrchars, (struct virterm *v, int r, int c, int n) );
#if PIXEL_WONDER
FORWARD _PROTOTYPE( void clrborder, (struct virterm *v)			);
#endif
FORWARD _PROTOTYPE( void lock_vduswitch, (struct tty_struct *tp)	);

/* Fast cursor positioning macroes for selected directions [see moveto()] */

#define	MOVERT(v)	if ((v)->ccol < NCOL-1) {			\
				(v)->ccol++;				\
				if ((v)->ccol & 1) (v)->curs++;	\
				else (v)->curs += increment_to_next_word; \
			} else						\
				if ((v)->wrap) (v)->ccol = NCOL;

#define	MOVELT(v)	if ((v)->ccol > 0) {				\
				if ((v)->ccol == NCOL) (v)->ccol--;	\
				(v)->ccol--;				\
				if ((v)->ccol & 1)			\
				   (v)->curs -= increment_to_next_word; \
				else (v)->curs--;			\
			}

#define	MOVEDN(v)	if (((v)->crow < (v)->botscroll) ||		\
			((v)->crow > (v)->botscroll && (v)->crow < (v)->nrow -1)){ \
				(v)->crow++; (v)->curs += (v)->bytr;	\
				if ((v)->ccol == NCOL) (v)->ccol--;	\
			}
#define	MOVEUP(v)	if (((v)->crow > (v)->topscroll) ||		\
			((v)->crow < (v)->topscroll && (v)->crow > 0)){ \
				(v)->crow--; (v)->curs -= (v)->bytr;	\
				if ((v)->ccol == NCOL) (v)->ccol--;	\
			}
#define	MOVECR(v)	{ if ((v)->ccol == NCOL) (v)->ccol--;		\
			  (v)->curs -= ((v)->ccol & ~1) *		\
					 (v)->nr_planes + ((v)->ccol&1);\
			(v)->ccol = 0; }

/*===========================================================================*
 *				flush					     *
 *===========================================================================*/
PUBLIC void flush(tp)
register struct tty_struct *tp;
{
/* There is really no semantic for flush() on 68000. The vdu driver does not
 * store characters in tp->tty_ramqueue, so simply return here.
 */

  return;
}

/*===========================================================================*
 *				console					     *
 *===========================================================================*/
PUBLIC void console(tp)
register struct tty_struct *tp;	/* tells which terminal is to be used */
{
/* Copy as much data as possible to the output queue, then start I/O.  On
 * memory-mapped terminals, such as the IBM console, the I/O will also be
 * finished, and the counts updated.  Keep repeating until all I/O done
 * or output is suspended.
 */

  register struct virterm *v = &virterm[tp->tty_line + NR_CONS];
  int count = 0;
  char *charptr = (char *)tp->tty_phys;

  vducursor(v, 0);
  while (tp->tty_outleft > 0 && tp->tty_inhibited == RUNNING) {
	vdu_char(tp, v, *charptr++);	/* write 1 byte to terminal */
	count++;
	tp->tty_outleft--;	/* decrement count */
  }
  vducursor(v, 1);

  /* Update terminal data structure. */
  tp->tty_phys += count;	/* advance physical data pointer */
  tp->tty_cum += count;		/* number of characters printed */

  /* If all data has been copied to the terminal, send the reply. */
  if (tp->tty_outleft == 0)
	finish(tp, tp->tty_cum);
}

/*===========================================================================*
 *				out_char				     *
 *===========================================================================*/
PUBLIC void out_char(tp, c)
register struct tty_struct *tp;	/* virtual terminal to use */
char c;				/* character to be output */
{
/* send character to VDU, with cursor control, for kernel putc() call.
 */

  register struct virterm *v = &virterm[tp->tty_line + NR_CONS];

  vducursor(v, 0);
  vdu_char(tp, v, c);
  vducursor(v, 1);
}

/*===========================================================================*
 *				vdu_char				     *
 *===========================================================================*/
PRIVATE void vdu_char(tp, v, c)
struct tty_struct *tp;		/* virtual terminal to use */
register struct virterm *v;	/* video descriptor for virt console */
register char c;		/* character to be output */
{
/* send character to VDU, collecting escape sequences
 */

  if (v->next == (char *)0) {
	if ((c & ~0x1F) == 0) {		/* control character */
		vductrl(tp, v, c);
		return;
	} else if (c != 0x7F) {
		if (c & 0x80)		/* map to font char 00 - 1F */
			c &= 0x1F;
		/* check if margin wrap is needed */
		if (v->ccol == NCOL && v->wrap == TRUE) {
			MOVECR(v);
			if (v->crow == v->botscroll)
				vscroll(v, FALSE, (v)->topscroll, v->botscroll, 1);
			else
				MOVEDN(v);
		}
		paint(v, c);	/* print normal char to screen */
		MOVERT(v);
		return;
	} else				/* do nothing on DEL */
		return;
  }
  if (v->next == v->vbuf && c == '[') {	/* ANSI esc sequence detected */
	*v->next++ = c;
	return;
  }
  if (c >= 060 && (v->next == v->vbuf || v->vbuf[0] != '[')) {
	vduesc(tp, v, c);		/* end of non-ANSI escape sequence */
	v->next = 0;
	return;
  }
  if (c >= 0100) {			/* end of ANSI escape sequence */
	vduansi(tp, v, c);
	v->next = 0;
	return;
  }
  if (c == '\030' || c == '\032') {	/* CAN or SUB cancels esc sequence */
	v->next = 0;
	/* no error char */
	return;
  }
  if ((c & ~0x1F) == 0) {	/* control character in an escape */
	vductrl(tp, v, c);	/* process it, do not save in vbuf */
	return;
  }
  if (v->next <= &v->vbuf[sizeof(v->vbuf)])
	*v->next++ = c;			/* record part of esc sequence */
}

/*
 * control character
 */
PRIVATE void vductrl(tp, v, c)
register struct tty_struct *tp;
register struct virterm *v;
char c;
{
  switch (c) {
  case 000: /* NUL: ignore */
  	return;
  case 005: /* ENQ: Transmit answerback */
  	return;
  case 007: /* BEL */
	beep();
	return;
  case 010: /* BS */
	MOVELT(v);
	return;
  case 011: /* HT */
	do {
		if ((tp->tty_mode & XTABS) == XTABS)
			paint(v, ' ');
		MOVERT(v);
	} while (v->ccol < NCOL-1 && (v->tabs[v->ccol] == 0));
	return;
  case 012: /* LF */
	if (!(tp->tty_mode & RAW) && (tp->tty_mode & CRMOD))
		MOVECR(v);
	/*FALLTHROUGH*/
  case 013: /* VT */
  case 014: /* FF */
	if (v->crow == v->botscroll)
		vscroll(v, FALSE, v->topscroll, v->botscroll, 1);
	else
		MOVEDN(v);
	return;
  case 015: /* CR */
	MOVECR(v);
	return;
  case 016: /* SO: select G1 character set */
  	return;
  case 017: /* SI: select G0 character set */
  	return;
  case 033: /* ESC */
	v->next = v->vbuf;
	return;
  default:
	return;
  }
}

/*
 * execute non-ANSI escape sequence
 */
PRIVATE void vduesc(tp, v, c)
struct tty_struct *tp;
register struct virterm *v;
char c;
{
  int qr, qc;
  
  if (v->next >= &v->vbuf[sizeof(v->vbuf)-1])
	return;
  *v->next = (char)c;
  switch (v->vbuf[0]) {
  case '8': /* DECRC: restore cursor */
	v->ccol = v->savccol;
	v->crow = v->savcrow;
	v->origmode = v->savorigmode;
	v->attr = v->savattr;
	moveto(v, v->crow, v->ccol);
	return;
  case '7': /* DECSC: save cursor */
	v->savccol = v->ccol;
	v->savcrow = v->crow;
	v->savorigmode = v->origmode;
	v->savattr = v->attr;
	return;
  case '=': /* DECKPAM: keypad application mode */
	v->vkeypad = TRUE;
	if (tp == cur_cons) keypad = TRUE;
	return;
  case '>': /* DECKPNM: keypad numeric mode */
	v->vkeypad = FALSE;
	if (tp == cur_cons) keypad = FALSE;
	return;
  case 'E': /* NEL: next line */
	MOVECR(v);
	/*FALLTHROUGH*/
  case 'D': /* IND: index */
	if (v->crow == v->botscroll)
		vscroll(v, FALSE, v->topscroll, v->botscroll, 1);
	else
		MOVEDN(v);
	return;
  case 'M': /* RI: reverse index */
	if (v->crow == v->topscroll)
		vscroll(v, TRUE, v->topscroll, v->botscroll, 1);
	else
		MOVEUP(v);
	return;
  case 'H': /* HTS: Hoizontal tab set */
  	v->tabs[v->ccol] = 1;
  	return;
  case 'c': /* RIS: reset to initial state */
	vduinit(tp);
	/* vduinit() calls vduswitch if neccessary */
	return;
  case '#': /* double height/width or alignment */
	switch (v->vbuf[1]) {
		case '8' :	/* DECALN - alignment screen */
			clrarea(v, 0, 0, v->nrow-1, NCOL-1);  /* clear screen */
			v->attr &= ~(DISPLAY_ATTRIBS);
			for (qr = 0; qr < v->nrow; qr++) {
				for (qc = 0; qc < NCOL; qc++) {
					moveto(v, qr, qc);
					paint(v, 'E');
				}
			}
			return;
		case '3' :	/* DECDHL - double height top */
		case '4' :	/* DECDHL - double height bottom */
		case '6' :	/* DECDWL - double width line */
		case '5' :	/* DECSWL - single width line */
		default:
			return;
	}
	return;
  case '(': /* SCS - select G0 character set */
	switch (v->vbuf[1]) {
		case 'A' :	/* UK char set as G0 */
		case 'B' :	/* US char set as G0 */
		case '0' :	/* Graphics char set as G0 */
		default:
			return;
	}
	return;
  case ')': /* SCS - select G1 character set */
	switch (v->vbuf[1]) {
		case 'A' :	/* UK char set as G1 */
		case 'B' :	/* US char set as G1 */
		case '0' :	/* Graphics char set as G1 */
		default:
			return;
	}
	return;
  case 'N': /* SS2 - select G2 char set for 1 character */
  case 'O': /* SS3 - select G3 char set for 1 character */
  default:
	return;
  }
}

/*
 * execute ANSI escape sequence
 */
PRIVATE void vduansi(tp, v, c)
struct tty_struct *tp;
register struct virterm *v;
char c;
{
  static unsigned colors[] = {		/* color table for ESC [ m */
	RGB_BLACK,
	RGB_RED,
	RGB_GREEN,
	RGB_YELLOW,
	RGB_BLUE,
	RGB_MAGENTA,
	RGB_CYAN,
	RGB_WHITE,
	RGB_LGREY,
	RGB_DGREY,
  };
  register int n, m, dec_private;	/* used for escape sequence params */
  unsigned char tmp_attr;
  
  if (v->next >= &v->vbuf[sizeof(v->vbuf)])
	return;
  *v->next = 0;				/* prepare for parameter scanning */
  dec_private = 0;
  v->next = &v->vbuf[1];
  if (v->vbuf[1] == '?') {
  	dec_private = 1;
  	v->next++;
  }
  n = vduparam(v);			/* all but one sequence use this */
  m = (n <= 0 ? 1 : n);

  switch (c) {
  case 'A': /* CUU: cursor up */
	while (m--) MOVEUP(v);
	return;
  case 'B': /* CUD: cursor down */
	while (m--) MOVEDN(v);
	return;
  case 'C': /* CUF: cursor forward */
	while (m--) MOVERT(v);
	return;
  case 'D': /* CUB: cursor backward */
	while (m--) MOVELT(v);
	return;
  case 'H': /* CUP: cursor position */
  case 'f': /* HVP: horizontal and vertical position */
	m = vduparam(v);
	n = (n <= 0 ? 1 : (n > v->nrow ? v->nrow : n)) - 1;
	m = (m <= 0 ? 1 : (m > NCOL    ? NCOL    : m)) - 1;
	if (v->origmode) {	/* adjust for origin mode */
		n = n + v->topscroll;
		if (n > v->botscroll)
			n = v->botscroll;
	}
	moveto(v, n, m);
	return;
  case 'r': /* DECSTBM: set top and bottom of scroll region */
	m = vduparam(v);
	n = (n <= 0 ? 1 : n) - 1;
	m = (m <= 0 ? v->nrow : m) - 1;
	if (n < m && m < v->nrow) { /* reasonable margins ? */
		v->topscroll = n;
		v->botscroll = m;
		moveto(v, v->topscroll, 0);
	} else {		/* no go - set margins to the whole screen */
		v->topscroll = 0;
		v->botscroll = v->nrow - 1;
	}
	return;
  case 'J': /* ED: erase in display */
	if	(n <= 0)   clrarea(v, v->crow, v->ccol, v->nrow-1, NCOL-1);
	else if (n == 1)   clrarea(v, 0, 0, v->crow, v->ccol);
	else if (n == 2)   clrarea(v, 0, 0, v->nrow-1, NCOL-1);
	return;
  case 'K': /* EL: erase in line */
	if	(n <= 0)   clrchars(v, v->crow, v->ccol, NCOL - v->ccol);
	else if (n == 1)   clrchars(v, v->crow, 0, v->ccol + 1);
	else if (n == 2)   clrchars(v, v->crow, 0, NCOL);
	return;
#ifdef NEEDED
  case 'n': /* DSR: device status report */
	if (m == 6) {
		kbdput('\033', tp->tty_line);
		kbdput('[', tp->tty_line);
		n = v->crow+1;
		if (n >= 10) kbdput(n/10, tp->tty_line);
		kbdput(n%10, tp->tty_line);
		kbdput(';', tp->tty_line);
		n = v->ccol+1;
		if (n >= 10) kbdput(n/10, tp->tty_line);
		kbdput(n%10, tp->tty_line);
		kbdput('R', tp->tty_line);
	}
	else if (m == 5)
	{
		kbdput('\033', tp->tty_line);
		kbdput('[', tp->tty_line);
		kbdput('0', tp->tty_line);
		kbdput('n', tp->tty_line);
	}
	return;
  case 'c': /* DA: device attributes */
	if (n <= 0)
		kbdput('\033', tp->tty_line);
		kbdput('[', tp->tty_line);
		kbdput('?', tp->tty_line);
		kbdput('1', tp->tty_line);
		kbdput(';', tp->tty_line);
		kbdput('2', tp->tty_line);
		kbdput('c', tp->tty_line);
	return;
#endif
  case 'm': /* SGR: set graphic rendition */
	do {
		if (n <= 0)		v->attr &= ~DISPLAY_ATTRIBS;
		else if (n == 1)	{ v->attr |= BOLD; v->attr &= ~LIGHT; }
		else if (n == 2)	{ v->attr |= LIGHT; v->attr &= ~BOLD; }
		else if (n == 4)	v->attr |= UNDERLINE;
		else if (n == 7)	v->attr |= REVERSE;
		else if (n == 21)	v->attr &= ~BOLD;
		else if (n == 22)	v->attr &= ~LIGHT;
		else if (n == 24)	v->attr &= ~UNDERLINE;
		else if (n == 27)	v->attr &= ~REVERSE;
		else if (n >= 30 && n <= 39)
		{
  			v->vrgb[(1 << v->nr_planes) -1] = colors[n - 30];
			lock_vduswitch(tp);
#if PIXEL_WONDER
			clrborder(v);
#endif
		}
		else if (n >= 40 && n <= 49)
		{
			v->vrgb[0] = colors[n - 40];
			lock_vduswitch(tp);
#if PIXEL_WONDER
			clrborder(v);
#endif
		}
	} while ((n = vduparam(v)) >= 0);
	return;
  case 'L': /* IL: insert line */
  	if (v->crow <= v->botscroll && v->crow >= v->topscroll)
		vscroll(v, TRUE, v->crow, v->botscroll, m);
	return;
  case 'M': /* DL: delete line */
  	if (v->crow <= v->botscroll && v->crow >= v->topscroll)
		vscroll(v, FALSE, v->crow, v->botscroll, m);
	return;
  case '@': /* ICH: insert char */
	hscroll(v, TRUE, v->crow, v->ccol, NCOL-1, m);
	return;
  case 'P': /* DCH: delete char */
	hscroll(v, FALSE, v->crow, v->ccol, NCOL-1, m);
	return;
  case 'g': /* TBC: Tabulation clear */
  	if (n <= 0)	v->tabs[v->ccol] = 0;	/* clear a single tab */
	/* clear all tabs */
  	else if (n == 3) for(m = 0; m < NCOL; m++) v->tabs[m] = 0;
  	return;
  case 'h': /* SM: set mode */
	if (dec_private) {
	    do {
	    	switch(n) {
	    	    case 1:	/* DECCKM - set cursor key application mode */
			v->vapmode = TRUE;
			if (tp == cur_cons) app_mode = TRUE;
			break;
		    case 3:	/* DECCOLM - set 132 column mode */
		    	break;
		    case 4:	/* DECSCLM - smooth scrolling */
		    	break;
		    case 5:	/* DECSCNM - reverse screen */
		    	v->vrgb[0] = DEFAULT_FORGRND;
  			v->vrgb[(1 << v->nr_planes) -1] = DEFAULT_BKGRND;
			lock_vduswitch(tp);
#if PIXEL_WONDER
			clrborder(v);
#endif
		    	break;
		    case 6:	/* DECOM - set origin mode in scroll region */
		    	v->origmode = TRUE;
			moveto(v, v->topscroll, 0);
		    	break;
		    case 7:	/* DECAWM - auto-wrap */
		    	v->wrap = TRUE;
		    	break;
		    case 8:	/* DECARM - auto-repeat */
		    	/* v->repeatmode = TRUE */
		    	break;
		    case 9:	/* DECINLM - interlace mode */
		    	break;
		    case 25:	/* DECTCEM - enable cursor */
		    	v->cursenable = TRUE;
			vducursor(v, 1);
			break;
		    default:
		    	break;
		}
	    } while ((n = vduparam(v)) >= 0);
	}
	return;

  case 'l': /* RM: reset mode */
	if (dec_private) {
	    do {
	    	switch(n) {
	    	    case 1:	/* DECCKM - set cursor key cursor mode */
			v->vapmode = FALSE;
			if (tp == cur_cons) app_mode = FALSE;
			break;
		    case 3:	/* DECCOLM - set 80 column mode */
		    	break;
		    case 4:	/* DECSCLM - jump scrolling */
		    	break;
		    case 5:	/* DECSCNM - normal screen */
		    	v->vrgb[0] = DEFAULT_BKGRND;
  			v->vrgb[(1 << v->nr_planes) -1] = DEFAULT_FORGRND;
	    		lock_vduswitch(tp);
#if PIXEL_WONDER
			clrborder(v);
#endif
		    	break;
		    case 6:	/* DECOM - reset origin mode */
		    	v->origmode = FALSE;
		    	moveto(v, 0, 0);
		    	break;
		    case 7:	/* DECAWM - auto-wrap */
		    	v->wrap = FALSE;
		    	break;
		    case 8:	/* DECARM - auto-repeat */
		    	/* v->repeatmode = FALSE */
		    	break;
		    case 9:	/* DECINLM - interlace mode */
		    	break;
		    case 25:	/* DECTCEM - disable cursor */
			vducursor(v, 0);
		    	v->cursenable = FALSE;
			break;
		    default:
		    	break;
		}
	    } while ((n = vduparam(v)) >= 0);
	}
	return;

  case '~': /* Minix-ST specific escape sequence */
	/*
	 * Handle the following escape sequence:
	 *   ESC [ l;a;m;r;g;b '~'
	 * where
	 *   if l present:
	 *	8: 8 line high font
	 *	12: 12 line high font
	 *	16: 16 line high font
	 *	24: 24 line screen (vt100 compatible)
	 *	25: 25 line screen (or more)
	 *   if a present:
	 *	attribute byte value (see paint())
	 *   if m present:
	 *	interpret r;g;b as colors for map register m
	 *	only assign color if r, g or b present
	 */
	if ((n == 8) || (n == 12) || (n == 16)) {
		v->linc = n;
		v->nrow = vdu_params[resolution].nr_video_lines / n;
		v->bytr = v->bytc * n;
		v->origmode = FALSE;
		v->topscroll = 0;
		v->botscroll = v->nrow - 1;
		moveto(v, v->nrow - 1, 0);
#if PIXEL_WONDER
		clrborder(v);
#endif
	}
	if (n == 24) {			/* 24 lines */
		v->nrow = 25;
		v->linc = vdu_params[resolution].nr_video_lines / v->nrow;
		if (v->linc >= 16) v->linc = 16;
		else if (v->linc >= 12) v->linc = 12;
		else v->linc = 8;
		v->bytr = v->bytc * v->linc;
		tmp_attr = v->attr;
		v->attr |= REVERSE;
		moveto(v, 24, 0);
		for (n = 0; n < NCOL; n++) {
			paint(v, statline[n]);
			MOVERT(v);
		}
		v->attr = tmp_attr;
		v->nrow = 24;
		v->origmode = FALSE;
		v->topscroll = 0;
		v->botscroll = v->nrow - 1;
		moveto(v, v->nrow - 1, 0);
#if PIXEL_WONDER
		clrborder(v);
#endif
	}
	if (n == 25) {			/* 25 lines (or more) */
		v->nrow = vdu_params[resolution].nr_video_lines / v->linc;
		v->bytr = v->bytc * v->linc;
		v->origmode = FALSE;
		v->topscroll = 0;
		v->botscroll = v->nrow - 1;
		clrchars(v, v->nrow - 1, 0, NCOL);		
		moveto(v, v->nrow - 1, 0);
	}
	if ((n = vduparam(v)) >= 0)	/* attribute byte */
		v->attr = n & 0xFF;
	if (((n = vduparam(v)) >= 0) &&
            (n < (1 << vdu_params[resolution].nr_planes)))
	{	/* change a color */
		register short *w = (short *)&v->vrgb[n & 0x0F];

		if ((n = vduparam(v)) >= 0)
			*w = (*w & ~0x0f00) | ((n & 0xf) << 8);
		if ((n = vduparam(v)) >= 0)
			*w = (*w & ~0x00f0) | ((n & 0xf) << 4);
		if ((n = vduparam(v)) >= 0)
			*w = (*w & ~0x000f) | ((n & 0xf) << 0);
		lock_vduswitch(tp);
#if PIXEL_WONDER
		clrborder(v);
#endif
	}
	return;
  default:
	return;
  }
}

/*
 * compute next parameter out of ANSI sequence
 */
PRIVATE int vduparam(v)
register struct virterm *v;
{
  register char c;
  register int i;

  i = -1;
  c = *v->next++;
  if (c >= '0' && c <= '9') {
	i = 0;
	do {
		i *= 10;
		i += (c - '0');
		c = *v->next++;
	} while (c >= '0' && c <= '9');
  }

  /* test for empty parameter (counts as zero) */
  if (i == -1 && c == ';') i = 0;
  
  if (c != ';')
	v->next--;
  return(i);
}

/*===========================================================================*
 *				manipulate videoram 			     *
 *===========================================================================*/
/* The routines in this part are all quite cryptic, but are optimized for max
 * speed. --kub
 */

/*
 * clear part of screen between two points inclusive
 */
PRIVATE void clrarea(v, rl, cl, ru, cu)
struct virterm *v;
int rl, cl;
int ru, cu;
{
  if (++cu == NCOL) {
	cu = 0;
	ru++;
  }
  if (cl > 0 && rl < ru) {
	clrchars(v, rl, cl, NCOL-cl);
	cl = 0;
	rl++;
  }
  clrlines(v, rl, ru-rl);
  clrchars(v, ru, cl, cu-cl);
}

/*
 * scroll lines around on the screen. newly created lines are blanked out
 */
PRIVATE void vscroll(v, up, rl, ru, n)
register struct virterm *v;
int up;				/* flag: TRUE = scroll up rather then down */
int rl, ru, n;			/* lower and upper row affected, scroll count */
{
  register char *src, *dst;	/* pointers into the video mem */

  if (n <= 0) return;		/* safety first ... */
  if (n > ru-rl)
	n = ru-rl + 1;

  if (up) {
  	src = v->vram + (ru-n) * (long)v->bytr;
  	dst = v->vram +  ru    * (long)v->bytr;
	for (; ru >= rl+n; ru--) {
		phys_copy((phys_bytes) src, (phys_bytes) dst,
			  (phys_bytes) v->bytr);
		src -= v->bytr; dst -= v->bytr;
	}
	clrlines(v, rl, n);
  } else {
	/* We are using phys_copy() to move all lines at once, but src and
	 * dest may overlap. Trust in phys_copy copying from low to high !
	 * Do not copy the last phys_byte using phys_copy since
	 * phys_copy may use movem and movem does one access beyond the
	 * end of the src dir, which may result in a bus error.
	 * clean way : for (; rl <= ru-n; rl++) phys_copy(...); (see above)
	 */
	register phys_bytes len = (phys_bytes) (ru-rl-n+1) * (long)v->bytr
				- sizeof(phys_bytes);
	src = v->vram + (rl+n) * (long)v->bytr;
	dst = v->vram +  rl    * (long)v->bytr;
	phys_copy((phys_bytes)src, (phys_bytes)dst, len);
	src += len;
	dst += len;
	*((phys_bytes *)dst) = *((phys_bytes *)src);
	clrlines(v, ru-n+1, n);
  }
}

/*
 * clear lines from r to r+n-1.
 */
PRIVATE void clrlines(v, r, n)
register struct virterm	*v;
int r, n;
{
  register phys_bytes	p, i;

  if (n <= 0) return;

  /* The ST video chips enforce the screen to be on a 256 byte boundary. As
   * long as CLICK_SIZE is smaller than that we can use zeroclicks for the
   * operation. Otherwise use a normal, but slow loop for clearing.
   */
  p = (phys_bytes) v->vram + r * (long)v->bytr;
  i = (phys_bytes) v->bytr * n;

  if ((p & (CLICK_SIZE-1)) == 0 && (i & (CLICK_SIZE-1)) == 0) {
	zeroclicks((phys_clicks) (p >> CLICK_SHIFT),
		   (phys_clicks) (i >> CLICK_SHIFT));
  } else {
	register long *q = (long *)p;

	i = i >> 4;	/* 16 bytes per loop */
	while (i--) {
		*q++ = 0; *q++ = 0; *q++ = 0; *q++ = 0;
	}
  }
}

#if PIXEL_WONDER
/*
 * Clear invisible border for PixelWonder.
 */
PRIVATE void clrborder(v)
register struct virterm	*v;
{
  register int i;
  register unsigned char *q;
  if (v->mono && v->bytc != NCOL) {
	for (q = v->vram, i = v->linc * v->nrow; i; q += v->bytc, i--)
		memset(q + NCOL, (v->vrgb[0] == RGB_WHITE)?0xff:0, v->bytc - NCOL);
  }
  /* What is to do for color??? */
}
#endif

/*
 * scroll chars around in one line. newly created chars are blanked out
 */
PRIVATE void hscroll(v, rt, r, cl, cu, n)
struct virterm *v;
int rt;				/* flag: TRUE = scroll right rather then left */
int r;				/* row number */
int cl, cu;			/* column scroll edges */
int n;				/* scroll count */
{
  char *src, *dst;		/* ptrs into video memory */
  register char *p, *q;		/* work pointers into vmem */
  register int nl;		/* line counter */
  register int bl = v->bytc;	/* line distance */
  register int pc;		/* plane counter */

  if (n <= 0) return;		/* safety checks */
  if (n > cu-cl)
	n = cu-cl + 1;

  if (rt) {
	src = v->vram + (r * (long)v->bytr) +
				((cu-n) & ~1) * v->nr_planes + ((cu-n) & 1);
	dst = v->vram + (r * (long)v->bytr) +
				((cu  ) & ~1) * v->nr_planes + ((cu  ) & 1);
	while (cu >= cl + n) {
		pc = v->nr_planes -1;
		do {
			p = src + 2 * pc;
			q = dst + 2 * pc;
			nl = v->linc;
			do {
				*q = *p;
				p += bl;
				q += bl;
			} while (--nl != 0);
		}
		while (pc--);
		if (!((long)src & 1)) src -= increment_to_next_word;
		else src--;
		if (!((long)dst & 1)) dst -= increment_to_next_word;
		else dst--;
		cu--;
	}
  } else {
	src = v->vram + (r * (long)v->bytr) +
				((cl+n) & ~1) * v->nr_planes + ((cl+n) & 1);
	dst = v->vram + (r * (long)v->bytr) +
				((cl  ) & ~1) * v->nr_planes + ((cl  ) & 1);
	while (cl <= cu - n) {
		pc = v->nr_planes - 1;
		do {
			p = src + 2 * pc;
			q = dst + 2 * pc;
			nl = v->linc;
			do {
				*q = *p;
				p += bl;
				q += bl;
			} while (--nl != 0);
		}
		while (pc--);
		if ((long)src & 1) src += increment_to_next_word;
		else src++;
		if ((long)dst & 1) dst += increment_to_next_word;
		else dst++;
		cl++;
	}
  }
  clrchars(v, r, cl, n);
}

/*
 * clear n chars on pos (r,c).
 */
PRIVATE void clrchars(v, r, c, n)
register struct virterm *v;
int r, c, n;
{
  register char *p, *q;		/* ptr into video memory */
  register int nl;		/* scan lines per char */
  register int bl = v->bytc;	/* bytes per scan line */
  register int pc;		/* plane counter */
           int planes = v->nr_planes;

  if (n > NCOL-c)
	n = NCOL-c;
  if (n <= 0) return;

  p = q = v->vram + (r * (long)v->bytr) + (c & ~1)* planes + (c & 1);
  while (n != 0) {
	if ((c & 1) || (n < sizeof(long))) {
		nl = v->linc;
		do {
			for (pc = planes-1; pc >= 0; pc--)
				*(p+2*pc) = 0;
			p += bl;
		} while (--nl != 0);
		if ((long)q & 1)	/* bit plane alignment */
			p = (q += increment_to_next_word);
		else	p = ++q;
		n--;
	}
	while (n >= sizeof(long)) {
		nl = v->linc;
		do {
			for (pc = planes-1; pc >= 0; pc--)
				*((long *)p+pc) = 0;
			p += bl;
		} while (--nl != 0);
		p = (q +=planes*sizeof(long));
		n -= sizeof(long);
	}
  }
}

/*===========================================================================*
 *				vducursor				     *
 *===========================================================================*/
PRIVATE void vducursor(v, onoff)
register struct virterm *v;
int	onoff;
{
/* (un-)display the cursor. onoff may eventually later have more semantic,
 * such as blinking. At the moment, (non)zero means cursor off(on).
 */

  register char *vp;
  register int nl;
  register int bl;
  register int planes = v->nr_planes;

  if (onoff == v->curmode) return;	/* no action if no change */
  if (v->cursenable == FALSE) return;  /* no action while cursor is inhibited */
  v->curmode = onoff;

  bl = v->bytc;
  do {
	planes--;
	nl = v->linc;
	vp = v->curs + 2*planes;
	do {
		*vp = ~(*vp); vp += bl;
	} while (--nl != 0);
  } while (planes);
}


/*===========================================================================*
 *				paint					     *
 *===========================================================================*/
PRIVATE void paint(v, c)
register struct virterm *v;
char c;
{
/* copy from font memory into video memory. */

  register unsigned char *fp;	/* ptr into font */
  register unsigned char *vp;	/* ptr into video memory */
  register int nl = v->linc;	/* scan lines per char */
  register int bl = v->bytc;	/* bytes per scan line */
  register int planes = v->nr_planes;
  register unsigned char attr = v->attr;
  c &= 0x7F;

  /* find the chararacter in the appropriate font */
  if (nl == 16)
	fp = &font16[c<<4];
  else
  if (nl == 12)
	fp = &font12[c * 12];
  else
	fp = &font8[c<<3];

  do {
	planes--;
	vp =  (unsigned char *)v->curs + 2*planes; /* write planes reversed */
	do {
	    if (attr & BOLD)
	    {
		*vp = *fp | (*fp++>>1);
	    }
	    else if (attr & LIGHT)
	    {
		if (nl & 1)
		{
		    *vp = (*fp | (*fp++>>1)) & 0xaa;
		}
		else
		{
		    *vp = (*fp | (*fp++>>1)) & 0x55;
		}
	    }
	    else
	    {
		*vp = *fp++;
	    }
	    if (attr & REVERSE) {
	        *vp = ~(*vp);
	    }
	    vp += bl;
	} while (--nl != 0);
	if (attr & UNDERLINE) {
		vp -= bl;
		*vp = (attr & REVERSE) ? 0x00 : 0xff;
	}
	nl = v->linc;
	fp -= nl;
  } while (planes);
}


/*===========================================================================*
 *				beep					     *
 *===========================================================================*/
PRIVATE void beep()
{
/* ring the console bell. This is independent from virtual consoles, just
 * because ringing the bell is an alarm signal for the user. He can then
 * switch to the appropriate virtual terminal and see what has happened.
 */

  register int	i, s;
  static char	sound[] = {
	0xA8,0x01,0xA9,0x01,0xAA,0x01,0x00,
	0xF8,0x10,0x10,0x10,0x00,0x20,0x03
  };

  s = lock();
  for (i = 0; i < sizeof(sound); i++) {
	SOUND->sd_selr = i;
	SOUND->sd_wdat = sound[i];
  }
  restore(s);
}

/*===========================================================================*
 *				moveto					     *
 *===========================================================================*/
PRIVATE void moveto(v, r, c)
register struct virterm *v;
int r, c;
{
/* move cursor to another position. The index operation to v->vram costs time
 * due to the multiplication, thus use the quick-and-dirty macros for simple
 * moves.
 */

  if (r < 0 || r >= v->nrow || c < 0 || c >= NCOL)
	return;
  if (v->origmode && (r < v->topscroll || r > v->botscroll))
  	return;
  v->crow = r;
  v->ccol = c;

  v->curs = v->vram + (r * (long)v->bytr) + (c & ~1) * v->nr_planes + (c & 1);
}

/*===========================================================================*
 *				lock_vduswitch				     *
 *===========================================================================*/
PRIVATE void lock_vduswitch(tp)
struct tty_struct *tp;
{
  /* safe gateway to vduswitch to block the  kbd interrupt handler to 
     avoid that during the execution of vduswitch the kbd interrrupt handler
     tries to do the same, and thus causes an inconsistent situation
     The kbd interrupt handler already has interrupts disabled and thus
     it calls vduswitch directly.
  */
  int s;

  s = lock();
  if (tp == cur_cons) vduswitch(tp);
  restore(s);
}

/*===========================================================================*
 *				vduswitch				     *
 *===========================================================================*/
PUBLIC void vduswitch(tp)
struct tty_struct *tp;
{
/* switch to another virtual console. The keyboard driver selects a new
 * virtual console by handling some special alt-key sequences. This procedure
 * is internally used to make changes to the current console visible. It's the
 * only place where keyboard variables and video controller are changed.
 */

  register struct virterm	*v = &virterm[tp->tty_line + NR_CONS];

  keypad = v->vkeypad;			/* keyboard initialization */
  app_mode = v->vapmode;
  increment_to_next_word = (v->nr_planes - 1) * 2 + 1;

  vdusetup((unsigned int)(v->vres), v->vram, v->vrgb);

  cur_cons = tp;
}

/*===========================================================================*
 *				vdusetup				     *
 *===========================================================================*/
PUBLIC void vdusetup(vres, vram, vrgb)
  unsigned int vres;
  char *vram;
  unsigned short *vrgb;
{
/* This routine prepares the video hardware programming. Doing it straight
 * away will often cause strange visual effects on the Mega-STe. It's safer
 * (and quite harmless for the other models) to wait until the next Vertical
 * Blank interrupt.
 */
  register int i;

  vid_res = vres;
  vid_ram = vram;

#if (ATARI_TYPE == TT)
  for (i = 255; i >= 0; i--)
#else
  for (i = 15; i >= 0; i--)
#endif
	vid_rgb [i] = vrgb [i];

  progvdu = 1;
}

/*===========================================================================*
 *				vbl					     *
 *===========================================================================*/
PUBLIC void vbl()
{
/* handle the Vertical-Blank interrupt. This routine should be called only
 * when needed; namely when the video hardware needs reprogramming.
 */
  register int i;

  VIDEO->vd_raml = (phys_bytes) vid_ram;	/* STe extra */
  VIDEO->vd_ramm = (phys_bytes) vid_ram >> 8; /* screen image address */
  VIDEO->vd_ramh = (phys_bytes) vid_ram >> 16;
#if (ATARI_TYPE == TT)
  VIDEO->vd_tt_res = vid_res;		/* screen resolution */
  for (i = 255; i >= 0; i--)
	VIDEO->vd_tt_rgb[i] = vid_rgb[i];
#else
  VIDEO->vd_st_res = vid_res;		/* screen resolution */
  for (i = 15; i >= 0; i--)
	VIDEO->vd_st_rgb[i] = vid_rgb[i];
#endif
#if (ATARI_TYPE != ST)
  VIDEO->vd_loff = 0;
  VIDEO->vd_ste_hscroll = 0;
#endif
}

/*===========================================================================*
 *				vduinit					     *
 *===========================================================================*/
PUBLIC void vduinit(tp)
struct tty_struct *tp;
{
/* Initialize a virtual console. This routine is not only called during system
 * startup, thus be careful with memory allocation. Don't allocate if that has
 * already been done. All other fields may be safely overwritten. Make changes
 * visible if they occur on the current virtual console.
 */

  register struct virterm	*v = &virterm[tp->tty_line + NR_CONS];
  unsigned long	p;
  int tabpos;
  unsigned short video_res;

#if (ATARI_TYPE == TT)
  video_res = VIDEO->vd_tt_res;
  resolution = (video_res >> 8) & 0x7;
  if (resolution == 0) {
	VIDEO->vd_tt_res = 0x400;
	video_res = 0x400;
	resolution = 4;
  }
#else
  video_res = VIDEO->vd_st_res & 0x3;
  resolution = video_res;
#endif

  switch(resolution) {
    case ST_RES_LOW:
  	resolution = ST_RES_MID;
	video_res = video_res + 1;
    case ST_RES_MID:
	v->linc = 8;
	break;
    case ST_RES_HIGH:
	v->linc = 16;
#if (ATARI_TYPE == TT)
	/* possibly duochrome mode; initialise colors 254 and 255 as well */
	v->vrgb[254] = RGB_BLACK;		/* background */
	v->vrgb[255] = RGB_WHITE;		/* foreground */
#endif
	break;
    case TT_RES_LOW:
  	resolution = TT_RES_MID;
	video_res = video_res - 0x300;
    case TT_RES_MID:
	v->linc = 16;
	break;
    case TT_RES_HIGH:
	v->linc = 16;
	break;
  }
  v->attr = 0;
  v->vres = video_res;
  v->nr_planes = vdu_params[resolution].nr_planes;
  v->bytc = vdu_params[resolution].nr_bytes_per_line * v->nr_planes;
  v->nrow = vdu_params[resolution].nr_video_lines / v->linc;
  v->bytr = v->bytc * v->linc;
  v->vrgb[(1 << v->nr_planes) -1] = RGB_WHITE;		/* foreground */
  v->vrgb[0] = RGB_BLACK;		/* background */

  if (v->vram == (char *)0) {		/* no image allocated yet */
	/* subtract video memory from physical top of memory */
	p = (unsigned long)(mem_base[0] + mem_size[0]) << CLICK_SHIFT;
	/* align on 256 byte boundary for ST */
	p = p & 0xffffff00;
	p -= VIDEORAM;
	mem_size[0] = (vir_bytes)(p - mem_base[0]) >> CLICK_SHIFT;
	v->vram = (char *)p;
#if PIXEL_WONDER
	clrborder(v);
#endif
  }
  v->vkeypad = FALSE;			/* default keyboard settings */
  v->vapmode = FALSE;
  /* set default tabs */
  for (tabpos = 0; tabpos < NCOL; tabpos++)
  	v->tabs[tabpos] = (tabpos & TAB_MASK) ? 0 : 1;
#if LINEWRAP
  v->wrap = TRUE;			/* line-wrap on  */
#else	/* LINEWRAP */
  v->wrap = FALSE;			/* line-wrap off */
#endif	/* LINEWRAP */
  v->topscroll = 0;			/* top of scroll region */
  v->botscroll = v->nrow - 1;		/* bottom of scroll region */
  v->origmode = FALSE;
  v->cursenable = TRUE;			/* start off with cursor on */
  v->curmode = 0;			/* clear screen wipes cursor */
  clrarea(v, 0, 0, v->nrow-1, NCOL-1);	/* clear the entire screen */
  moveto(v, 0, 0);			/* move cursor to upper left corner */
  lock_vduswitch(tp);			/* initialize vdu controller */
}

/*===========================================================================*
 *				vdu_loadfont				     *
 *===========================================================================*/
PUBLIC int vdu_loadfont(m_ptr)
message *m_ptr;
{
    register phys_bytes	address;
    struct fnthdr *hdr;
    phys_bytes size;

    address = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS,
		   sizeof(struct fnthdr));
    hdr = (struct fnthdr *)address;
    size = (phys_bytes) (128 / 8) * hdr->width * hdr->heigth; /* table size */
    if (hdr->width != 8) return(EINVAL);
    address = numap(m_ptr->PROC_NR, (vir_bytes) hdr->addr, size);
    switch (hdr->heigth) {
 	case 8:
		phys_copy((phys_bytes)address, (phys_bytes)font8, size);
		break;
 	case 12:
		phys_copy((phys_bytes)address, (phys_bytes)font12, size);
		break;
 	case 16:
		phys_copy((phys_bytes)address, (phys_bytes)font16, size);
		break;
	default:
		return(EINVAL);
    }
    return(OK);
}
#endif
